/******************************************************************************
** Societe      : 4JS
** Programme    : mkchartar.c
** Date         : 26/04/1996
** Version      : 1.00
**
** Titre        : Make characters tables for f4gl compiler
******************************************************************************/

/******************************************************************************
**			     INCLUSIONS + MACROS
******************************************************************************/

#include<stdio.h>

#undef NULL
#define NULL	0

#define MAX_STR  512
#define MAX_CHAR 256
#define ISDIGIT(c) (c >= '0' && c <= '9' )


/***************************************************************************
** 			   GLOBALS VARIABLES
****************************************************************************/

char *progname;

/***************************************************************************
** 			 	FONCTIONS
****************************************************************************/

/******************************************************************************
** Usage of programme
** Parametre :
**	mess : Message
******************************************************************************/

usage(mess)

char *mess;
{
 if(mess!=NULL && *mess)
   fprintf(stderr, "%s : %s\n\n",progname,mess);

  fprintf(stderr,"Make characters tables for f4gl compiler.\n");
  fprintf(stderr,"Usage : %s fileframe [ctfile]\n",progname);
  fprintf(stderr,"  where    fileframe : Contains the IO map\n");
  fprintf(stderr,"           ctfile    : result file\n");
 exit(1);
}

/*******************************************************************************
** FUNCTION convert_inchar : Put 'x' , 0xdd, or xxx in char
** Parameters :
**  str   : String to be analysed
** Returnings :
**  value : Value or -1 if problem
*******************************************************************************/

short convert_inchar(str)

char *str;

{
 short value=(-1);
 long ind=1;

if ( strlen(str) == 3 && str[0] == '\'' && str[2] == '\'')
  value = str[1];
else if ( strlen(str) == 4 && !strncmp(str,"0x",2))
 {
  str+=3;
  value=0;
  while ( str[0] != 'x' )
   {
    if ( str[0] >= '0' && str[0] <= '9' )
      value += ind*(str[0] - '0' );
    else if ( str[0] >= 'A' && str[0] <= 'F' )
      value += ind*(10 + str[0] - 'A' );
    else if ( str[0] >= 'a' && str[0] <= 'f' )
      value += ind*(10 + str[0] - 'a' );
    else
      return(NULL);
    ind *= 16;
    str--;
   }
 }
else if ( strlen(str) == 3
      && ISDIGIT(str[0]) && ISDIGIT(str[1]) && ISDIGIT(str[2]) )
  value = atoi(str);

 return(value);

} /* END FUNCTION convert_inchar */

/*******************************************************************************
** FUNCTION read_mapfile : Read and verify the map file
** Parameters :
**  filename : Name of map file
**  input    : Array for the input map
**  output   : Array for the output map
** Returnings :
**  none
*******************************************************************************/

void read_mapfile(filename,input,output)

char *filename;
unsigned char *input;
unsigned char *output;

{

 short m_input[MAX_CHAR], m_output[MAX_CHAR], *curr_array, curr_char[2];
		/* Temporary storage of the mapping */

 char line[MAX_STR+1]; /* Current line of the file */
 char mess[MAX_STR+1]; /* Some error message */
 char path[MAX_STR+1]; /* Path of map file */
 register char *cpt;   /* Index for line */
 char *beg;
 register int i;
 long lineno=0;	       /* Number of line in the file */
 FILE *f;
 short like_io=0;	/* 1 = input like opposate output
			   2 = output like opposate input */

/***************************
* Try to open the map file *
***************************/
 if ( (f = fopen(filename,"r")) == NULL )
  {
   sprintf(mess,"File %s not found or not readable.",filename);
   usage(mess);
  }

 for (i=0; i < MAX_CHAR; i++)
  {
   m_input[i] = (-1);
   m_output[i] = (-1);
  }

 curr_array=NULL;

/****************************
* Read and analyse the file *
****************************/
 while( fgets(line,MAX_STR,f) != NULL)
  {
   lineno++;
   /**********************
   * Delete the final \n *
   **********************/
   if ( line[strlen(line)-1] ==  '\n' )
     line[strlen(line)-1]=0;
   cpt=line;

/*******************************
* Go after blank or tabulation *
*******************************/
   while (cpt[0] == ' ' || cpt[0] == '\t')
     cpt++;

/********************************************
* If the line isn't a comment or blank line *
********************************************/
   if(cpt[0] != '#' && cpt[0])
   /*****************************
   * Perhaps a part declaration *
   *****************************/
     if ( !strcmp(cpt,"[input]") || !strcmp(cpt,"input") )
        curr_array = m_input;
      else if ( !strcmp(cpt,"[output]") || !strcmp(cpt,"output"))
        curr_array = m_output;
     else if ( curr_array == NULL )
      {
       sprintf(mess,
	"In the map file line %d, you must say which part do you are.",lineno);
       usage(mess);
      }
     else
      {
      /***************************************
      *  This will be a map declaration line *
      ***************************************/
       beg=cpt;
       while (cpt[0] && cpt[0] != ' ' && cpt[0] != '\t')
         cpt++;
       if ( !cpt[0] )
        {
         sprintf(mess, "In the map file line %d, syntax error.",lineno);
         usage(mess);
        }
       cpt[0]=0;
       /******************************
       * Take the next [^ \t][^ \t]* *
       ******************************/
       curr_char[0]=convert_inchar(beg);
       if (curr_char[0] == (-1) || curr_char[0] >= MAX_CHAR)
        {
         sprintf(mess, "In the map file line %d, error with the first part of line.",lineno);
         usage(mess);
        }
       /******************************
       * Take the next [^ \t][^ \t]* *
       ******************************/
       cpt++;
       while (cpt[0] == ' ' || cpt[0] == '\t')
         cpt++;
       beg=cpt;
       while (cpt[0] && cpt[0] != ' ' && cpt[0] != '\t')
         cpt++;
       cpt[0]=0;
       curr_char[1]=convert_inchar(beg);
       if (curr_char[1] == (-1)  || curr_char[1] >= MAX_CHAR)
        {
         sprintf(mess, "In the map file line %d, error with the second part of line.",lineno);
         usage(mess);
        }
       for (i=0; i < 2; i++)
       curr_array[curr_char[0]]=curr_char[1];
      }
      /*******************************************************************
      * Declare than input (or output) are opposate of output (or input) *
      *******************************************************************/
   else if ( !strcmp(cpt,"#output") || !strcmp(cpt,"#tuptuo"))
     like_io = 1;
   else if ( !strcmp(cpt,"#input") || !strcmp(cpt,"#tupni"))
     like_io = 2;
  }
 fclose(f);

 /***************************
 * Proceed input and output *
 ***************************/
 /*for ( i = 0; i < MAX_CHAR; i++)
  {
   printf("%x,output=%x,input=%x\n",i,m_output[i],m_input[i]);
  }*/

 for ( i = 0; i < MAX_CHAR; i++)
  {
   if( m_input[i] == (-1))
     input[i] = (unsigned char)i;
   else
     input[i] = (unsigned char)m_input[i];
   /*printf("input[%x] =%x\n",i,input[i]);*/
  }

 for ( i = 0; i < MAX_CHAR; i++)
  {
   if( m_output[i] == (-1))
     output[i] = (unsigned char)i;
   else
     output[i] = (unsigned char)m_output[i];
   /*printf("output[%x] =%x\n",i,output[i]);*/
  }

 /*****************************
 * Input like opposate Output *
 *****************************/
 if (like_io == 1)
   for ( i = 0; i < MAX_CHAR; i++)
     input[output[i]] = (unsigned char )i;
 /*****************************
 * Output like opposate Input *
 *****************************/
 else if (like_io == 2)
   for ( i = 0; i < MAX_CHAR; i++)
     output[input[i]] = (unsigned char )i;

} /* END FUNCTION read_mapfile */


/******************************************************************************
**			       MAIN SECTION
******************************************************************************/

main(argc,argv)
int argc;
char **argv;
{
 unsigned char input[MAX_CHAR];		/* Input map array */
 unsigned char output[MAX_CHAR];	/* Output map array */
 int fout;
 char mess[100];
 
 progname = argv[0];

/*************************
* Analyse the parameters *
*************************/
 if (argc > 1 && ( !strcmp(argv[1], "help") || !strcmp(argv[1],"?")) )
   usage(NULL);

 if (argc > 1 && !strcmp(argv[1], "-v") )
   {
    fprintf(stderr,"%s Version 1.00.1a\n",progname);
    exit(1);
   }

 if (argc < 2 || argc > 3)
   usage(NULL);

/********************
* Read the Map file *
********************/

 read_mapfile(argv[1],input,output);

 if ( argc == 3 ) {
   if ( (fout = creat(argv[2],0777)) == (-1) )
    {
     sprintf(mess,"Cannot create file %s.\n",argv[2]);
     usage(mess);
    }
 } else {
   fout=1;
 }

 write(fout,input,MAX_CHAR);
 write(fout,output,MAX_CHAR);

 return(0);
}
